# Copyright (c) 2008-2019 by Enthought, Inc.
# All rights reserved.

from __future__ import absolute_import

try:
    from traits.version import version as __version__
except ImportError:
    # If we get here, we're using a source tree that hasn't been created via
    # the setup script. That likely also means that the ctraits extension
    # hasn't been built, so this isn't a viable Traits installation. OTOH, it
    # can be useful if a simple "import traits" doesn't actually fail.
    __version__ = "unknown"

# Add a NullHandler so 'traits' loggers don't complain when they get used.
import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())

del logging



# ------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   06/21/2002
#
#  Refactored into a separate module: 07/04/2003
#
# ------------------------------------------------------------------------------

""" Defines common, low-level capabilities needed by the Traits package.
"""

# -------------------------------------------------------------------------------
#  Imports:
# -------------------------------------------------------------------------------

from __future__ import absolute_import

import os
import sys
from os import getcwd
from os.path import dirname, exists, join

import six

from . import _py2to3
from ._py2to3 import LONG_TYPE

from .etsconfig.api import ETSConfig

# backwards compatibility: trait_base used to provide a patched enumerate
enumerate = enumerate

# -------------------------------------------------------------------------------
#  Constants:
# -------------------------------------------------------------------------------

ClassTypes = _py2to3.ClassTypes

SequenceTypes = (list, tuple)

ComplexTypes = (float, int)

TypeTypes = (
    str,
    six.text_type,
    int,
    LONG_TYPE,
    float,
    complex,
    list,
    tuple,
    dict,
    bool,
)

TraitNotifier = "__trait_notifier__"

# The standard Traits property cache prefix:
TraitsCache = "_traits_cache_"

# -------------------------------------------------------------------------------
#  Singleton 'Uninitialized' object:
# -------------------------------------------------------------------------------
Uninitialized = None


class _Uninitialized(object):
    """ The singleton value of this class represents the uninitialized state
        of a trait and is specified as the 'old' value in the trait change
        notification that occurs when the value of a trait is read before being
        set.
    """

    def __new__(cls):
        if Uninitialized is not None:
            return Uninitialized
        else:
            self = object.__new__(cls)
            return self

    def __repr__(self):
        return "<uninitialized>"

    def __reduce_ex__(self, protocol):
        return (_Uninitialized, ())


#: When the first reference to a trait is a 'get' reference, the default value of
#: the trait is implicitly assigned and returned as the value of the trait.
#: Because of this implicit assignment, a trait change notification is
#: generated with the Uninitialized object as the 'old' value of the trait, and
#: the default trait value as the 'new' value. This allows other parts of the
#: traits package to recognize the assignment as the implicit default value
#: assignment, and treat it specially.
Uninitialized = _Uninitialized()

# -------------------------------------------------------------------------------
#  Singleton 'Undefined' object (used as undefined trait name and/or value):
# -------------------------------------------------------------------------------

Undefined = None


class _Undefined(object):
    """ Singleton 'Undefined' object (used as undefined trait name and/or value)
    """

    def __new__(cls):
        if Undefined is not None:
            return Undefined
        else:
            self = object.__new__(cls)
            return self

    def __repr__(self):
        return "<undefined>"

    def __reduce_ex__(self, protocol):
        return (_Undefined, ())

    def __eq__(self, other):
        return type(self) is type(other)

    def __hash__(self):
        return hash(type(self))

    def __ne__(self, other):
        return type(self) is not type(other)


#: Singleton object that indicates that a trait attribute has not yet had a
#: value set (i.e., its value is undefined). This object is used instead of
#: None, because None often has other meanings, such as that a value is not
#: used. When a trait attribute is first assigned a value, and its associated
#: trait notification handlers are called, Undefined is passed as the *old*
#: parameter, to indicate that the attribute previously had no value.
Undefined = _Undefined()

# Tell the C-base code about singleton 'Undefined' and 'Uninitialized' objects:
from . import ctraits

ctraits._undefined(Undefined, Uninitialized)

# -------------------------------------------------------------------------------
#  Singleton 'Missing' object (used as missing method argument marker):
# -------------------------------------------------------------------------------


class Missing(object):
    """ Singleton 'Missing' object (used as missing method argument marker).
    """

    def __repr__(self):
        return "<missing>"


#: Singleton object that indicates that a method argument is missing from a
#: type-checked method signature.
Missing = Missing()

# -------------------------------------------------------------------------------
#  Singleton 'Self' object (used as object reference to current 'object'):
# -------------------------------------------------------------------------------


class Self(object):
    """ Singleton 'Self' object (used as object reference to current 'object').
    """

    def __repr__(self):
        return "<self>"


#: Singleton object that references the current 'object'.
Self = Self()

# -------------------------------------------------------------------------------
#  Define a special 'string' coercion function:
# -------------------------------------------------------------------------------


def strx(arg):
    """ Wraps the built-in str() function to raise a TypeError if the
    argument is not of a type in StringTypes.
    """
    if isinstance(arg, StringTypes):
        return str(arg)
    raise TypeError


# -------------------------------------------------------------------------------
#  Constants:
# -------------------------------------------------------------------------------

StringTypes = (str, six.text_type, int, LONG_TYPE, float, complex)

# -------------------------------------------------------------------------------
#  Define a mapping of coercable types:
# -------------------------------------------------------------------------------

# Mapping of coercable types.
CoercableTypes = {
    LONG_TYPE: (11, LONG_TYPE, int),
    float: (11, float, int),
    complex: (11, complex, float, int),
    six.text_type: (11, six.text_type, str),
}

# -------------------------------------------------------------------------------
#  Return a string containing the class name of an object with the correct
#  article (a or an) preceding it (e.g. 'an Image', 'a PlotValue'):
# -------------------------------------------------------------------------------


def class_of(object):
    """ Returns a string containing the class name of an object with the
    correct indefinite article ('a' or 'an') preceding it (e.g., 'an Image',
    'a PlotValue').
    """
    if isinstance(object, six.string_types):
        return add_article(object)

    return add_article(object.__class__.__name__)


# -------------------------------------------------------------------------------
#  Return a string containing the right article (i.e. 'a' or 'an') prefixed to
#  a specified string:
# -------------------------------------------------------------------------------


def add_article(name):
    """ Returns a string containing the correct indefinite article ('a' or 'an')
    prefixed to the specified string.
    """
    if name[:1].lower() in "aeiou":
        return "an " + name

    return "a " + name


# ----------------------------------------------------------------------------
#  Return a 'user-friendly' name for a specified trait:
# ----------------------------------------------------------------------------


def user_name_for(name):
    """ Returns a "user-friendly" version of a string, with the first letter
    capitalized and with underscore characters replaced by spaces. For example,
    ``user_name_for('user_name_for')`` returns ``'User name for'``.
    """
    name = name.replace("_", " ")
    result = ""
    last_lower = False

    for c in name:
        if c.isupper() and last_lower:
            result += " "
        last_lower = c.islower()
        result += c

    return result.capitalize()


# -------------------------------------------------------------------------------
#  Gets the path to the traits home directory:
# -------------------------------------------------------------------------------

_traits_home = None


def traits_home():
    """ Gets the path to the Traits home directory.
    """
    global _traits_home

    if _traits_home is None:
        _traits_home = verify_path(join(ETSConfig.application_data, "traits"))

    return _traits_home


# -------------------------------------------------------------------------------
#  Verify that a specified path exists, and try to create it if it doesn't:
# -------------------------------------------------------------------------------


def verify_path(path):
    """ Verify that a specified path exists, and try to create it if it
        does not exist.
    """
    if not exists(path):
        try:
            os.mkdir(path)
        except:
            pass

    return path


# -------------------------------------------------------------------------------
#  Returns the name of the module the caller's caller is located in:
# -------------------------------------------------------------------------------


def get_module_name(level=2):
    """ Returns the name of the module that the caller's caller is located in.
    """
    return sys._getframe(level).f_globals.get("__name__", "__main__")


# -------------------------------------------------------------------------------
#  Returns a resource path calculated from the caller's stack:
# -------------------------------------------------------------------------------


def get_resource_path(level=2):
    """Returns a resource path calculated from the caller's stack.
    """
    module = sys._getframe(level).f_globals.get("__name__", "__main__")

    path = None
    if module != "__main__":
        # Return the path to the module:
        try:
            path = dirname(getattr(sys.modules.get(module), "__file__"))
        except:
            # Apparently 'module' is not a registered module...treat it like
            # '__main__':
            pass

    if path is None:
        # '__main__' is not a real module, so we need a work around:
        for path in [dirname(sys.argv[0]), getcwd()]:
            if exists(path):
                break

    # Handle application bundlers.  Since the python source files may be placed
    # in a zip file and therefore won't be directly accessable using standard
    # open/read commands, the app bundlers will look for resources (i.e.  data
    # files, images, etc.) in specific locations.  For py2app, this is in the
    # [myapp].app/Contents/Resources directory.  For py2exe, this is the same
    # directory as the [myapp].exe executable file generated by py2exe.  For
    # pyinstaller, the attribute sys._MEIPASS is set to this directory.
    frozen = getattr(sys, "frozen", False)
    if frozen:
        if hasattr(sys, "_MEIPASS"):
            root = sys._MEIPASS
        elif frozen == "macosx_app":
            root = os.environ["RESOURCEPATH"]
        elif frozen in ("dll", "windows_exe", "console_exe"):
            root = os.path.dirname(sys.executable)
        else:
            # Unknown app bundler, but try anyway
            root = os.path.dirname(sys.executable)
        if ".zip/" in path:
            zippath, image_path = path.split(".zip/")
            path = os.path.join(root, image_path)

    return path


# -------------------------------------------------------------------------------
#  Returns the value of an extended object attribute name of the form:
#  name[.name2[.name3...]]:
# -------------------------------------------------------------------------------


def xgetattr(object, xname, default=Undefined):
    """ Returns the value of an extended object attribute name of the form:
        name[.name2[.name3...]].
    """
    names = xname.split(".")
    for name in names[:-1]:
        if default is Undefined:
            object = getattr(object, name)
        else:
            object = getattr(object, name, None)
            if object is None:
                return default

    if default is Undefined:
        return getattr(object, names[-1])

    return getattr(object, names[-1], default)


# -------------------------------------------------------------------------------
#  Sets the value of an extended object attribute name of the form:
#  name[.name2[.name3...]]:
# -------------------------------------------------------------------------------


def xsetattr(object, xname, value):
    """ Sets the value of an extended object attribute name of the form:
        name[.name2[.name3...]].
    """
    names = xname.split(".")
    for name in names[:-1]:
        object = getattr(object, name)

    setattr(object, names[-1], value)


# -------------------------------------------------------------------------------
#  Traits metadata selection functions:
# -------------------------------------------------------------------------------


def is_none(value):
    return value is None


def not_none(value):
    return value is not None


def not_false(value):
    return value is not False


def not_event(value):
    return value != "event"


def is_str(value):
    return isinstance(value, six.string_types)



# -----------------------------------------------------------------------------
#
#  Copyright (c) 2015, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in /LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
# -----------------------------------------------------------------------------
"""
Tests for the Float trait type.

"""
import unittest

import six

from traits.api import BaseFloat, Either, Float, HasTraits, TraitError, Unicode
from traits.testing.optional_dependencies import (
    numpy, requires_numpy, requires_python2)

if six.PY2:
    LONG_TYPE = long
else:
    LONG_TYPE = int


class MyFloat(object):
    def __init__(self, value):
        self._value = value

    def __float__(self):
        return self._value


class BadFloat(object):
    def __float__(self):
        raise ZeroDivisionError


class FloatModel(HasTraits):
    value = Float

    # Assignment to the `Either` trait exercises a different C code path (see
    # validate_trait_complex in ctraits.c).
    value_or_none = Either(None, Float)

    float_or_text = Either(Float, Unicode)


class BaseFloatModel(HasTraits):
    value = BaseFloat

    value_or_none = Either(None, BaseFloat)

    float_or_text = Either(Float, Unicode)


class CommonFloatTests(object):
    """ Common tests for Float and BaseFloat """

    def test_default(self):
        a = self.test_class()
        self.assertEqual(a.value, 0.0)

    def test_accepts_float(self):
        a = self.test_class()

        a.value = 5.6
        self.assertIs(type(a.value), float)
        self.assertEqual(a.value, 5.6)

        a.value_or_none = 5.6
        self.assertIs(type(a.value_or_none), float)
        self.assertEqual(a.value_or_none, 5.6)

    def test_accepts_int(self):
        a = self.test_class()

        a.value = 2
        self.assertIs(type(a.value), float)
        self.assertEqual(a.value, 2.0)

        a.value_or_none = 2
        self.assertIs(type(a.value_or_none), float)
        self.assertEqual(a.value_or_none, 2.0)

    def test_accepts_float_like(self):
        a = self.test_class()

        a.value = MyFloat(1729.0)
        self.assertIs(type(a.value), float)
        self.assertEqual(a.value, 1729.0)

        a.value = MyFloat(594.0)
        self.assertIs(type(a.value), float)
        self.assertEqual(a.value, 594.0)

    def test_rejects_string(self):
        a = self.test_class()
        with self.assertRaises(TraitError):
            a.value = "2.3"
        with self.assertRaises(TraitError):
            a.value_or_none = "2.3"

    def test_bad_float_exceptions_propagated(self):
        a = self.test_class()
        with self.assertRaises(ZeroDivisionError):
            a.value = BadFloat()

    def test_compound_trait_float_conversion_fail(self):
        # Check that a failure to convert to float doesn't terminate
        # an assignment to a compound trait.
        a = self.test_class()
        a.float_or_text = u"not a float"
        self.assertEqual(a.float_or_text, u"not a float")

    @requires_python2
    def test_accepts_small_long(self):
        a = self.test_class()
        a.value = LONG_TYPE(2)
        self.assertIs(type(a.value), float)
        self.assertEqual(a.value, 2.0)

        a.value_or_none = LONG_TYPE(2)
        self.assertIs(type(a.value_or_none), float)
        self.assertEqual(a.value_or_none, 2.0)

    @requires_python2
    def test_accepts_large_long(self):
        a = self.test_class()

        # Value large enough to be a long on Python 2.
        a.value = 2 ** 64
        self.assertIs(type(a.value), float)
        self.assertEqual(a.value, 2 ** 64)

        a.value_or_none = 2 ** 64
        self.assertIs(type(a.value_or_none), float)
        self.assertEqual(a.value_or_none, 2 ** 64)

    @requires_numpy
    def test_accepts_numpy_floats(self):
        test_values = [
            numpy.float64(2.3),
            numpy.float32(3.7),
            numpy.float16(1.28),
        ]
        a = self.test_class()
        for test_value in test_values:
            a.value = test_value
            self.assertIs(type(a.value), float)
            self.assertEqual(a.value, test_value)

            a.value_or_none = test_value
            self.assertIs(type(a.value_or_none), float)
            self.assertEqual(a.value_or_none, test_value)


class TestFloat(unittest.TestCase, CommonFloatTests):
    def setUp(self):
        self.test_class = FloatModel

    def test_exceptions_propagate_in_compound_trait(self):
        # This test doesn't currently pass for BaseFloat, which is why it's not
        # in the common tests. That's probably a bug.
        a = self.test_class()
        with self.assertRaises(ZeroDivisionError):
            a.value_or_none = BadFloat()


class TestBaseFloat(unittest.TestCase, CommonFloatTests):
    def setUp(self):
        self.test_class = BaseFloatModel



# ------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   11/06/2004
#
# ------------------------------------------------------------------------------

""" Adds a "category" capability to Traits-based classes, similar to that
    provided by the Cocoa (Objective-C) environment for the Macintosh.

    You can use categories to extend an existing HasTraits class, as an
    alternative to subclassing. An advantage of categories over subclassing is
    that you can access the added members on instances of the original class,
    without having to change them to instances of a subclass. Unlike
    subclassing, categories do not allow overriding trait attributes.
"""

# -------------------------------------------------------------------------------
#  Imports:
# -------------------------------------------------------------------------------

from __future__ import absolute_import

import warnings

import six

from .has_traits import (
    MetaHasTraits,
    update_traits_class_dict,
    BaseTraits,
    ClassTraits,
    InstanceTraits,
    PrefixTraits,
    ListenerTraits,
    ViewTraits,
)

# -------------------------------------------------------------------------------
#  'MetaCategory' class:
# -------------------------------------------------------------------------------


class MetaCategory(MetaHasTraits):
    """
    Metaclass providing magic for the category extension mechanism.

    .. deprecated:: 5.2
       The category extension mechanism is deprecated, and the Category
       and MetaCategory classes will be removed in a future version of Traits.

    """
    def __new__(cls, class_name, bases, class_dict):

        # Make sure the correct usage is being applied:
        if len(bases) > 2:
            raise TypeError(
                "Correct usage is: class FooCategory(Category,Foo):"
            )

        # Categories are deprecated. Only warn if len(bases) == 2 to avoid
        # a spurious warning when creating the Category class itself.
        if len(bases) == 2:
            # See enthought/traits#319 for deprecation rationale.
            warnings.warn(
                (
                    "Use of the Category class is deprecated. Category "
                    "will be removed in a future version of Traits."
                ),
                DeprecationWarning,
                stacklevel=2,
            )

        # Process any traits-related information in the class dictionary:
        update_traits_class_dict(
            class_name, bases, class_dict, is_category=True
        )

        if len(bases) == 2:
            category_class = bases[1]

            # Update the class and each of the existing subclasses:
            category_class._add_trait_category(
                class_dict.pop(BaseTraits),
                class_dict.pop(ClassTraits),
                class_dict.pop(InstanceTraits),
                class_dict.pop(PrefixTraits),
                class_dict.pop(ListenerTraits),
                class_dict.pop(ViewTraits),
            )

            # Move all remaining items in our class dictionary to the base
            # class's dictionary:
            for name, value in list(class_dict.items()):
                if not hasattr(category_class, name):
                    setattr(category_class, name, value)
                    del class_dict[name]

        # Finish building the class using the updated class dictionary:
        return type.__new__(cls, class_name, bases, class_dict)


# -------------------------------------------------------------------------------
#  'Category' class:
# -------------------------------------------------------------------------------


@six.add_metaclass(MetaCategory)
class Category(object):
    """ Used for defining "category" extensions to existing classes.

    .. deprecated:: 5.2
       The category extension mechanism is deprecated, and the Category
       and MetaCategory classes will be removed in a future version of Traits.

    To define a class as a category, specify "Category," followed by the name
    of the base class name in the base class list.

    The following example demonstrates defining a category::

        from traits.api import HasTraits, Str, Category

        class Base(HasTraits):
            x = Str("Base x")
            y = Str("Base y")

        class BaseExtra(Category, Base):
            z = Str("BaseExtra z")
    """



from .adapter import Adapter, PurePythonAdapter

from .adaptation_error import AdaptationError

from .adaptation_manager import (
    adapt,
    AdaptationManager,
    get_global_adaptation_manager,
    provides_protocol,
    register_factory,
    register_offer,
    register_provides,
    reset_global_adaptation_manager,
    set_global_adaptation_manager,
    supports_protocol,
)

from .adaptation_offer import AdaptationOffer



# ------------------------------------------------------------------------------
# Copyright (c) 2013, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# ------------------------------------------------------------------------------
""" Base classes for adapters.

Adapters do not have to inherit from these classes, as long as their
constructor takes the object to be adapted as the first and only
*positional* argument.

"""


from traits.has_traits import HasTraits
from traits.trait_types import Any


class PurePythonAdapter(object):
    """ Base class for pure Python adapters. """

    def __init__(self, adaptee):
        """ Constructor. """

        self.adaptee = adaptee

        return


class Adapter(HasTraits):
    """ Base class for adapters with traits. """

    def __init__(self, adaptee, **traits):
        """ Constructor. """

        traits["adaptee"] = adaptee
        super(Adapter, self).__init__(**traits)

        return

    adaptee = Any



""" Examples for lazy adapter factories.

This module should be only imported when the adaptation takes place.
"""


class IBar(object):
    pass


class IBarToIFoo(object):
    pass


class IFoo(object):
    pass



# -----------------------------------------------------------------------------
#
#  Copyright (c) 2019, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in /LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
# -----------------------------------------------------------------------------

import unittest

from traits.testing.optional_dependencies import optional_import


class TestImportHandler(unittest.TestCase):
    def test_import_succeeds(self):

        module = optional_import("itertools")
        self.assertEqual(module.__name__, "itertools")

    def test_import_fails(self):

        module = optional_import("unavailable_module")
        self.assertIsNone(module)
